// oheap.cpp
//

#include "oheap.h"
#include "vheap.h"

#include <Psapi.h>

vheap gvheap;

oheap* oheap::get()
{
    static oheap heap;
    return &heap;
}

oheap::oheap() :
    sleepCount(0),
    stopMallocs(false),
    runThreads(true),
    mallocfct(&oheap::speedymalloc),
    thread(&oheap::start)
{
}

oheap::~oheap()
{
    runThreads = false;
    thread.join();
}

void oheap::start()
{
    get()->oheap_();
}

void oheap::oheap_()
{
    mallocfct_t speedyMalloc = &oheap::speedymalloc;
    mallocfct_t sloppyMalloc = &oheap::sloppymalloc;
    mallocfct_t stopMalloc = &oheap::stopmalloc;

    const SIZE_T oneGigabyte = 1024 * 1024 * 1024;

    PROCESS_MEMORY_COUNTERS_EX pmc;

    const HANDLE currentProcess = GetCurrentProcess();

    GetProcessMemoryInfo(currentProcess, (PROCESS_MEMORY_COUNTERS*)&pmc, sizeof(pmc));

    SIZE_T pageFileUsagePrev = pmc.PageFaultCount;

    MEMORYSTATUSEX status;
    status.dwLength = sizeof(status);

    while(runThreads)
    {
        int errorCode = GlobalMemoryStatusEx(&status);
        if (errorCode == 0)
        {
            exit(1);
        }

        stopMallocs.store( status.ullAvailPhys < oneGigabyte );

        if (stopMallocs)
        {
            mallocfct.store( stopMalloc );

            //gvheap.vunlock();     VirtualUnlock on the mmaped regions, this call shows that re-allocated pages are not flushed back to disk
            //SetProcessWorkingSetSize(currentProcess, -1, -1);    Forces pages in workspace to be flushed
            EmptyWorkingSet(currentProcess); // Flush unused pages from workspace
            
            sleepCount.store(100);
            mallocfct.store( sloppyMalloc  );
        }
        else if(sleepCount>0)
            sleepCount.fetch_sub(1);
        else
            mallocfct.store( speedyMalloc );

        Sleep(1);
    }
}

void* oheap::sloppymalloc(size_t size_p)
{
    Sleep(sleepCount);
    return gvheap.allocateNextMemBlock(size_p);
}

void* oheap::stopmalloc(size_t size_p)
{
    while (stopMallocs)
        Sleep(0);
    return gvheap.allocateNextMemBlock(size_p);
}

void* oheap::speedymalloc(size_t size_p)
{
    return gvheap.allocateNextMemBlock(size_p);
}

void* oheap::malloc(size_t size_p)
{
    return ((*this).*(mallocfct_t)mallocfct)(size_p);
}

void oheap::free(void* p)
{
    char* p_ = (char*)p;
    gvheap.freeMemBlock((__s_block*)(p_ - BLOCK_SIZE));
}

