// Dr. Thomas Chust
// https://chust.org/repos/mmu
// memmap.h
#pragma once

#include <Windows.h>

#include <cstddef>
#include <string>
#include <memory>

/// Memory mapping utilities.
namespace memmap {
    /// Options for memory mappings.
    struct Options {
        bool m_ReadOnly : 1; ///< Whether the mapping is read-only.
        bool m_Create : 1;   ///< Whether the file may be created.

        /// Initialize option defaults.
        /// The read-only flag is true by default.
        /// The create flag is set iff the read-only flag is not set.
        explicit Options(const bool readonly_p = true) : m_ReadOnly(readonly_p), m_Create(!readonly_p) {}
    };

    class MMapBuffer
    {
        HANDLE m_File, m_Mapping;
        void* m_Data;
        size_t m_Begin, m_Size, m_Page;

        static void Size2DWORDs(DWORD& high_p, DWORD& low_p, const size_t size_p);

    public:
        static size_t PageSize();
        static void ThrowLastError();
        MMapBuffer(){}
        explicit MMapBuffer(const char* path_p, const size_t begin_p, const size_t size_p, const Options options_p);
        ~MMapBuffer() /*override*/;
        void* data() const /*override*/;
        size_t size() const /*override*/;
    };

};


